#include <ncurses.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <ctype.h>
#include "common.h"
#include "proto.h"

#define UPDATE_DELAY 2000000

char map[MAP_YSIZE][MAP_XSIZE];
char special[MAP_YSIZE][MAP_XSIZE];
char current_map[128];
int p_x;
int p_y;
int diamonds_left;
int lives;
long int score;
long int score_last_extralife;
int level;
int custom_map;
int need_refresh;
int option_sound;

int main(int argc, char **argv)
{
  chk_all();

  curses_start();

  if(COLS < 80 || LINES < 25)
    bail("error: your terminal size must be at least 80x25");

  signal(SIGINT, sigint_handler);
  signal(SIGWINCH, sigwinch_handler);

  srand(time(0));

  make_ready();

  option_sound = 1;

  if(argv[1]) {
    snprintf(current_map, sizeof current_map - 1, "%s", argv[1]);
    custom_map = 1;
  }

  else {
    current_map[0] = 0x00;
    custom_map = 0;
    splash();
  }
 
  mainloop();

  return EXIT_SUCCESS;
}


void make_ready(void)
{
  level = 1;
  lives = 3;
  score = 0;
  score_last_extralife = 0;
  need_refresh = 0;
}

int mainloop(void)
{
  int i;
  int input;
  int old_p_x;
  int old_p_y;
  int x_direction = 0;
  int update_delay;
  int changes;
  int x, y;
  int rval;

  erase();

  create_map(current_map);
  fix_map();

  p_x = 1;
  p_y = 1;

  for(y = 0; y < MAP_YSIZE; y++) {
    for(x = 0; x < MAP_XSIZE; x++) {
      if(map[y][x] == MAP_PLAYER) {
        p_x = x;
        p_y = y;
      }
    }
  }

  map[p_y][p_x] = MAP_PLAYER;

  diamonds_left = count_diamonds();

  while(update_map() > 0);
  draw_map();
  draw_status();
  refresh();

  flushinp();

  while(1) {

    if(need_refresh == 1) {
      draw_map();
      draw_status();
      refresh();
      need_refresh = 0;
    }

    if(diamonds_left <= 0) {
      level_done(p_x, p_y);
    }

    if(isready(0)) {

      input = getch();
      flushinp();

      if(tolower(input) == 'q') {
        for(;;) {
          rval = tolower(msgbox("Are you sure you want to quit? (Yes/No)"));
          if(rval == 'y' || rval == '\n' || rval == ' ') {
            curses_stop();
            exit(0);
          }
          else if(rval == 'n') {
            break;
          }
        }
      }

      if(tolower(input) == 'k') {
        map[p_y][p_x] = MAP_STONE;
        special[p_y][p_x] = SPECIAL_DIAMOND;
        player_died();
      }
 
      if(tolower(input) == 's') {
        if(option_sound == 0) {
          option_sound = 1;
          _beep();
        }
        else option_sound = 0;
      }

      if(tolower(input) == 'w') {  /* stupid feature to make lh stfu */
 
         for(i = 0; i < 3; i++) {
           attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
           mvaddch(p_y-1, p_x, '+');
           mvaddch(p_y+3, p_x, '+');
           mvaddch(p_y+1, p_x+2, '+');
           mvaddch(p_y+1, p_x-2, '+');
           attrset(A_NORMAL);
           refresh();
           usleep(90000);
           draw_map();
           refresh();
           usleep(50000);
        }

        erase();
        draw_map();
        draw_status();
        refresh();

      }

      map[p_y][p_x] = 0;
      old_p_y = p_y;
      old_p_x = p_x;

      if(input == KEY_UP || input == '8')    { p_y--; x_direction = +0; }
      if(input == KEY_DOWN || input == '2')  { p_y++; x_direction = +0; }
      if(input == KEY_LEFT || input == '4')  { p_x--; x_direction = -1; }
      if(input == KEY_RIGHT || input == '6') { p_x++; x_direction = +1; }

      if(map[p_y][p_x] == MAP_WALL) {
        p_y = old_p_y;
        p_x = old_p_x;
      }

      if(map[p_y][p_x] == MAP_STONE && x_direction == -1 && map[p_y][p_x-1] == MAP_EMPTY) {
        if(special[p_y][p_x] == 1) {
          special[p_y][p_x] = 0;
          map[p_y][p_x] = MAP_EMPTY;
          got_diamond();         
        }
        else if(special[p_y][p_x] == 2) {
          special[p_y][p_x] = 0;
          map[p_y][p_x] = MAP_EMPTY;
          got_money();         
        }
        else {
          map[p_y][p_x-1] = MAP_STONE;
        }

      }

      else if(map[p_y][p_x] == MAP_STONE && x_direction == +1 && map[p_y][p_x+1] == MAP_EMPTY) {
        if(special[p_y][p_x] == 1) {
          special[p_y][p_x] = 0;
          map[p_y][p_x] = MAP_EMPTY;
          got_diamond();
        }
        else if(special[p_y][p_x] == 2) {
          special[p_y][p_x] = 0;
          map[p_y][p_x] = MAP_EMPTY;
          got_money();
        }
        else {
          map[p_y][p_x+1] = MAP_STONE;
        }
      }

      else if(map[p_y][p_x] == MAP_STONE && special[p_y][p_x] == 0) {
        p_y = old_p_y;
        p_x = old_p_x;
      }

      else if(map[p_y][p_x] == MAP_STONE && special[p_y][p_x] == 1) {
        special[p_y][p_x] = 0;
        got_diamond();
      }

      else if(map[p_y][p_x] == MAP_STONE && special[p_y][p_x] == 2) {
        special[p_y][p_x] = 0;
        got_money();
      }
     
      map[p_y][p_x] = MAP_PLAYER;

      update_delay = UPDATE_DELAY;

      for(;;) {
        draw_map();
        draw_status();
        refresh();

        for(i = 0; i < update_delay; i++)
          continue;

        changes = update_map();
        if(changes == 0) {
          break;
        }

        if(update_delay > 0) {
          update_delay = update_delay - (UPDATE_DELAY / 50);
        }

        if(isready(0)) {
          break;
        }

      }

    }

    else {
      usleep(1000);
    }

  }

  curses_stop();
  return EXIT_SUCCESS;
}


void draw_map(void)
{
  int x, y;

  for(y = 0; y < MAP_YSIZE; y++) {
    for(x = 0; x < MAP_XSIZE; x++) {
      if(map[y][x] == MAP_EMPTY)   mvaddch(y + 1, x, CHR_EMPTY);
      if(map[y][x] == MAP_DIRT)    mvaddch(y + 1, x, CHR_DIRT);
      if(map[y][x] == MAP_WALL)    mvaddch(y + 1, x, CHR_WALL);
      if(map[y][x] == MAP_PLAYER)  mvaddch(y + 1, x, CHR_PLAYER);

      if(map[y][x] == MAP_STONE)   mvaddch(y + 1, x, CHR_STONE);
      if(map[y][x] == MAP_STONE && special[y][x] == 1) 
         mvaddch(y + 1, x, CHR_DIAMOND);
      if(map[y][x] == MAP_STONE && special[y][x] == 2) 
         mvaddch(y + 1, x, CHR_MONEY);

    }
  }

}


int update_map(void)
{
  int x, y;
  int changes = 0;

  for(y = 0; y < MAP_YSIZE; y++) {
    for(x = 0; x < MAP_XSIZE; x++) {

      if(map[y][x] == MAP_STONE && map[y+1][x] == MAP_EMPTY) {
        map[y][x] = MAP_EMPTY;
        map[y+1][x] = MAP_STONE;
        if(special[y][x] == 1) {
          special[y][x] = 0;
          special[y+1][x] = 1;
        }
        if(special[y][x] == 2) {
          special[y][x] = 0;
          special[y+1][x] = 2;
        }

        if(map[y+1][x] == MAP_STONE && map[y+2][x] == MAP_PLAYER) {
          player_died();
        }

        return 1;
      }

      if(map[y][x] == MAP_STONE && map[y+1][x] == MAP_STONE && map[y+2][x] != MAP_EMPTY) {
        if(map[y][x-1] == MAP_EMPTY && map[y+1][x-1] == MAP_EMPTY) {
          map[y][x] = MAP_EMPTY;
          map[y+1][x-1] = MAP_STONE;

          if(special[y][x] == 1) {
            special[y][x] = 0;
            special[y+1][x-1] = 1;
          }

          if(special[y][x] == 2) {
            special[y][x] = 0;
            special[y+1][x-1] = 2;
          }

          if(map[y+1][x-1] == MAP_STONE && map[y+2][x-1] == MAP_PLAYER) {
            player_died();
          }

          return 1;
        }
        if(map[y][x+1] == MAP_EMPTY && map[y+1][x+1] == MAP_EMPTY) {
          map[y][x] = MAP_EMPTY;
          map[y+1][x+1] = MAP_STONE;

          if(special[y][x] == 1) {
            special[y][x] = 0;
            special[y+1][x+1] = 1;
          }

          if(special[y][x] == 2) {
            special[y][x] = 0;
            special[y+1][x+1] = 2;
          }

          if(map[y+1][x+1] == MAP_STONE && map[y+2][x+1] == MAP_PLAYER) {
            player_died();
          }

          return 1;
        }
      }
    }
  }

  return changes;
}


void create_map(char *mapname)
{
  int y, x;
  char mstr[64];

  for(y = 0; y < MAP_YSIZE; y++) {
    for(x = 0; x < MAP_XSIZE; x++) {
      map[y][x] = MAP_EMPTY;
      special[y][x] = 0;
    }
  }

  if(mapname[0] == 0x00) {
    snprintf(mstr, sizeof mstr - 1, "data/levels/%02d", level);
  }

  else {
    snprintf(mstr, sizeof mstr - 1, "%s", mapname);
  }

  if(load_map(mstr, map) == 1) {
    bail("error: load_map failed");
  }

  while(update_map() != 0);

}


void player_died(void)
{
  int rval;

  lives--;

  update_map();
  draw_map();
  draw_status();
  refresh();  

  _beep();
  usleep(90000);
  explode(p_x, p_y, 4, EX_C);

  while(update_map() > 0) {
    _beep();
    draw_map();
    refresh();
  }
 
  _beep();
  
  sleep(1);

  if(lives <= 0) {

    gameover();
    sleep(2);

    for(;;) {

      rval = tolower(msgbox("Game over! Play again? (Y/N)"));

      if(rval == 'y' || rval == '\n' || rval == ' ') {
        make_ready();
        if(custom_map == 0) {
          current_map[0] = 0x00;
        }
        mainloop();
      }

      else if(rval == 'n') {
        curses_stop();
        exit(0);
      }

    }  

  } else {

    sleep(1);
    mainloop();

  }
}


void explode(int x, int y, int len, int chr)
{
  int offset;

  explode_put(y, x, chr);
  map[y][x] = MAP_EMPTY;

  refresh();

  for(offset = 0; offset < len; offset++) {
    explode_put(y + offset, x, chr);
    explode_put(y - offset, x, chr);
    explode_put(y, x + offset * 2, chr);
    explode_put(y, x - offset * 2, chr);

    explode_put(y + offset, x + offset, chr);
    explode_put(y - offset, x + offset, chr);
    explode_put(y - offset, x - offset, chr);
    explode_put(y + offset, x - offset, chr);

    refresh();
    usleep(EX_DELAY);
  }

}


void explode_put(int y, int x, int chr)
{
  if((x > 1 && MAP_XSIZE - 2 > x) && (y > 1 && MAP_YSIZE - 1> y)) {
    mvaddch(y, x, chr);
    map[y][x] = MAP_STONE;
    special[y][x] = 1;
  }
}


int count_diamonds()
{
  int x, y;
  int num_diamonds = 0;

  for(y = 0; y < MAP_YSIZE; y++) {
    for(x = 0; x < MAP_XSIZE; x++) {
      if(special[y][x] == 1) {
        num_diamonds++;
      }
    }
  }

  return num_diamonds;
}


void got_diamond()
{
  _beep();
  diamonds_left--;
  score += POINTS_DIAMOND;
  if(score >= score_last_extralife + 1000)
    got_extralife();
}


void got_money()
{
  _beep();
  score += POINTS_MONEY;
  if(score >= score_last_extralife + 1000)
    got_extralife();
}


void got_extralife()
{
  int i;

  if(lives < 99) {

    for(i = 0; i < 6; i++) {
      _beep();
      usleep(1);
    }

    lives++;
    score_last_extralife = score;
    draw_status();

  }
}


void fix_map(void)
{
  int x, y;

  for(y = 0; y < MAP_YSIZE; y++) {
    for(x = 0; x < MAP_XSIZE; x++) {
      if(map[y][x] == MAP_DIAMOND) {
        map[y][x] = MAP_STONE;
        special[y][x] = 1;
      }
      if(map[y][x] == MAP_MONEY) {
        map[y][x] = MAP_STONE;
        special[y][x] = 2;
      }
    }
  }
 
}


void level_done(int x, int y)
{
  int i;
  usleep(100000);
  mvaddch(y + 1, x, 'z');
  refresh();
  usleep(200000);
  mvaddch(y + 1, x, '.');
  refresh();
  usleep(200000);

  map[y][x] = MAP_EMPTY;

  for(i = 6; i < 12; i += 2) {
    explode(MAP_XSIZE / 2, MAP_YSIZE / 2, i, EX_C);
    while(update_map() > 0) {
      _beep();
      draw_map();
      refresh();
    }
  }

  sleep(1);

  level++;
  mainloop();
}


void draw_status(void)
{
  attrset(COLOR_PAIR(COLOR_GREEN));
  mvprintw(0, calc_center(strlen("CAVEZ of PHEAR ("VERSION")")), "CAVEZ of PHEAR ("VERSION")");

  attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
  mvprintw(0, 0, "DIAMONDS LEFT:");
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  mvprintw(0, 15, "%02d", diamonds_left);

  attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
  mvprintw(0, 71, "LIVES:");
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  mvprintw(0, 78, "%02d", lives);

  attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
  mvprintw(24, 0, "SCORE:");
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  mvprintw(24, 7, "%d", score);

  attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
  mvprintw(24, 71, "LEVEL:");
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  mvprintw(24, 78, "%02d", level);

  attrset(A_NORMAL);
}

void _beep(void)
{
  if(option_sound == 1) {
    beep();
  }
}
